﻿using SoftwareInstallationBusinessLogic.BusinessLogics;
using SoftwareInstallationContracts.BindingModels;
using SoftwareInstallationContracts.BusinessLogicsContracts;
using SoftwareInstallationContracts.SearchModels;
using SoftwareInstallationContracts.StoragesContracts;
using SoftwareInstallationContracts.ViewModels;
using SoftwareInstallationDataModels.Models;
using Microsoft.Extensions.Logging;

namespace SoftwareInstallationBusinessLogic
{
    public class ShopLogic : IShopLogic
    { 
        private readonly ILogger _logger;
        private readonly IShopStorage _shopStorage;
        public ShopLogic(ILogger<ShopLogic> logger, IShopStorage shopStorage)
        {
            _logger = logger;
            _shopStorage = shopStorage;
        }
        public List<ShopViewModel>? ReadList(ShopSearchModel? model)
        {
            _logger.LogInformation("ReadList. ShopName:{ShopName}.Id:{ Id} ",
                model?.Name, model?.Id);
            var list = (model == null) ? _shopStorage.GetFullList() :
            _shopStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count:{Count}", list.Count);
            return list;
        }
        public ShopViewModel? ReadElement(ShopSearchModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            _logger.LogInformation("ReadElement. ShopName:{ShopName}.Id:{ Id}",
                model.Name, model.Id);
            var element = _shopStorage.GetElement(model);
            if (element == null)
            {
                _logger.LogWarning("ReadElement element not found");
                return null;
            }
            _logger.LogInformation("ReadElement find. Id:{Id}", element.Id);
            return element;
        }
        public bool Create(ShopBindingModel model)
        {
            CheckModel(model);
            model.Packages = new();
            if (_shopStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }
            return true;
        }
        public bool Update(ShopBindingModel model)
        {
            CheckModel(model, false);
            if (string.IsNullOrEmpty(model.Name))
            {
                throw new ArgumentNullException("Нет названия магазина",
                nameof(model.Name));
            }
            if (_shopStorage.Update(model) == null)
            {
                _logger.LogWarning("Update operation failed");
                return false;
            }
            return true;
        }
        public bool Delete(ShopBindingModel model)
        {
            CheckModel(model, false);
            _logger.LogInformation("Delete. Id:{Id}", model.Id);
            if (_shopStorage.Delete(model) == null)
            {
                _logger.LogWarning("Delete operation failed");
                return false;
            }
            return true;
        }
        private void CheckModel(ShopBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (!withParams)
            {
                return;
            }
            if (string.IsNullOrEmpty(model.Name))
            {
                throw new ArgumentNullException("Нет названия магазина",
                nameof(model.Name));
            }
            if (model.MaxCountPackages < 0)
            {
                throw new ArgumentException(
                    "Максимальное количество изделий в магазине не должно быть отрицательным",
                    nameof(model.MaxCountPackages));
            }
            _logger.LogInformation("Shop. ShopName:{0}.Address:{1}. Id: {2}",
                model.Name, model.Address, model.Id);
            var element = _shopStorage.GetElement(new ShopSearchModel
            {
                Name = model.Name
            });
            if (element != null && element.Id != model.Id && element.Name == model.Name)
            {
                throw new InvalidOperationException("Магазин с таким названием уже есть");
            }
        }

        public bool AddPackage(ShopSearchModel model, IPackageModel package, int count)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (count <= 0)
            {
                throw new ArgumentException("Количество добавляемого изделия должно быть больше 0", nameof(count));
            }
            
            _logger.LogInformation("AddPackageInShop. ShopName:{ShopName}.Id:{ Id}",
                model.Name, model.Id);
            var element = _shopStorage.GetElement(model);
            int currCount = element.Packages.Values.First().Item2;
            if (element == null)
            {
                _logger.LogWarning("AddPackageInShop element not found");
                return false;
            }
            if (count + currCount > element.MaxCountPackages)
            {
                throw new ArgumentException("Количество изделий не должно быть больше максимального количества изделий");
                return false;
            }
           
            _logger.LogInformation("AddPackageInShop find. Id:{Id}", element.Id);

            var prevCount = element.Packages.GetValueOrDefault(package.Id, (package, 0)).Item2;
            element.Packages[package.Id] = (package, prevCount + count);
            _logger.LogInformation(
                    "AddPackageInShop. Has been added {count} {package} in {ShopName}",
                    count, package.PackageName, element.Name);

            _shopStorage.Update(new()
            {
                Id = element.Id,
                Address = element.Address,
                Name = element.Name,
                DateOpening = element.DateOpening,
                Packages = element.Packages
            });
            return true;
        }
        public int GetFreePlacesWithPackagesInShops(int countPackages)
        {
            // Сумма разностей между максимальный кол-вом изделий и суммой всех изделий в магазине
            return _shopStorage.GetFullList()
                .Select(x => x.MaxCountPackages - x.Packages
                .Select(p => p.Value.Item2).Sum())
                .Sum() - countPackages;
        }

        public bool AddPackagesInShops(IPackageModel package, int count)
        {
            if (count <= 0)
            {
                _logger.LogWarning("AddPackagesInShops. Количество добавляемых изделий должно быть больше 0. Количество - {count}", count);
                return false;
            }
            var freePlaces = GetFreePlacesWithPackagesInShops(count);
            if (freePlaces < 0)
            {
                _logger.LogInformation("AddPackagesInShops. Не удалось добавить изделия в магазины, поскольку они переполнены." +
                    "Освободите магазины на {places} изделий", -freePlaces);
                return false;
            }
            foreach (var shop in _shopStorage.GetFullList())
            {
                var cnt = Math.Min(count, shop.MaxCountPackages - shop.Packages.Select(x => x.Value.Item2).Sum());
                if (cnt <= 0)
                {
                    continue;
                }
                if (!AddPackage(new() { Id = shop.Id }, package, cnt))
                {
                    _logger.LogWarning("При добавления изделий во все магазины произошла ошибка");
                    return false;
                }
                count -= cnt;
                if (count == 0)
                {
                    return true;
                }
            }
            return true;
        }

        public bool SellPackages(IPackageModel package, int needCount)
        {
            return _shopStorage.SellPackages(package, needCount);
        }
    }
}